! ------------------------------------------------------------------------
!>  \brief Module for testing Constitutive Matrix module
! ------------------------------------------------------------------------
MODULE constitutive_test
USE fruit                   !> Unit testing framework
USE system_constants        !> Global system constants
USE log_message_control     !> Printing log/error messages
USE log_messages            !> Log/error message and sender codes
USE dense_matrix_def        !> Dense Matrix ADT
USE constitutive            !> Constitutive Matrix module

IMPLICIT NONE


CONTAINS


! ------------------------------------------------------------------------
!>  \test Test for EXCEED exception from dmatrix
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param emod       Non-exceeding elastic modulus
!!  \param nu         Non-exceeding Poisson's ratio
!!  \param emod1      Test elastic modulus 1 ( = E_MIN-0.1 < E_MIN )
!!  \param nu1        Test Poisson's ratio 1 ( = NU_MIN-0.1 < NU_MIN )
!!  \param emod1      Test elastic modulus 2 ( = EMAX+1.0 > E_MAX )
!!  \param nu1        Test Poisson's ratio 2 ( = 0.5 > NU_MAX )
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param Dmat       Dummy matrix for dmatrix output
!!
!!  This test attempts to obtain the constitutive matrix for values that
!!  are outside the specified range and ensures that the correct exception
!!  is raised.
! ------------------------------------------------------------------------
SUBROUTINE test_constitutive_EXCEED
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_constitutive_EXCEED'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  DOUBLE PRECISION, PARAMETER :: emod = 70.d3, nu = 0.3d0
  DOUBLE PRECISION, PARAMETER :: emod1 = E_MIN-0.1d0
  DOUBLE PRECISION, PARAMETER :: nu1 = NU_MIN-0.1d0
  DOUBLE PRECISION, PARAMETER :: emod2 = E_MAX+1.d0
  DOUBLE PRECISION, PARAMETER :: nu2 = 0.5d0
  INTEGER, PARAMETER :: expMsg=EXCEED
  INTEGER :: actMsg
  TYPE(matrixT) :: Dmat

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log message file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> try to call with low elastic modulus
  CALL dmatrix(emod1,nu, Dmat, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to call with high elastic modulus
  CALL dmatrix(emod2,nu, Dmat, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to call with low Poisson's ratio
  CALL dmatrix(emod,nu1, Dmat, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to call with high Poisson's ratio
  CALL dmatrix(emod,nu2, Dmat, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate D matrix
  CALL log_closeLogFile()
  CALL dm_clean(Dmat)

END SUBROUTINE test_constitutive_EXCEED


! ------------------------------------------------------------------------
!>  \test Test for dmatrix when elastic modulus is zero
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param emod       Elastic modulus ( = 0 )
!!  \param nu         Poisson's ratio
!!  \param expDat     Expected constitutive matrix (zeros)
!!  \param Dmat       Matrix for dmatrix output
!!
!!  This test obtains the constitutive matrix when elastic modulus is
!!  zero and Poisson's ratio is any value within the system defined
!!  limits. It should be all zeros.
! ------------------------------------------------------------------------
SUBROUTINE test_constitutive_ZERO
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_constitutive_ZERO'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  DOUBLE PRECISION, PARAMETER :: emod = 0.d0
  DOUBLE PRECISION, PARAMETER :: nu = 0.3d0
  DOUBLE PRECISION, DIMENSION(NTNS,NTNS) :: expDat
  TYPE(matrixT) :: Dmat

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected result
  expDat = RESHAPE( (/ &
                    0.d0, 0.d0, 0.d0, &
                    0.d0, 0.d0, 0.d0, &
                    0.d0, 0.d0, 0.d0  &
                    /), SHAPE(expDat) )

  !> get constitutive matrix
  CALL dmatrix(emod,nu, Dmat)
  CALL assertEquals(expDat, Dmat%dat, NTNS,NTNS)

  !> deallocate constitutive matrix
  CALL dm_clean(Dmat)

END SUBROUTINE test_constitutive_ZERO


! ------------------------------------------------------------------------
!>  \test Test for dmatrix when parameters are in normal range
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param emod       Elastic modulus
!!  \param nu         Poisson's ratio
!!  \param toler      Tolerance on output
!!  \param expDat     Expected constitutive matrix
!!  \param Dmat       Matrix for dmatrix output
!!
!!  This test obtains the constitutive matrix when elastic modulus
!!  and Poisson's ratio are given values within the system defined
!!  limits.
! ------------------------------------------------------------------------
SUBROUTINE test_constitutive_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_constitutive_VAL'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  DOUBLE PRECISION, PARAMETER :: emod = 70.d3
  DOUBLE PRECISION, PARAMETER :: nu = 0.3d0
  DOUBLE PRECISION, PARAMETER :: toler = 1.d-9
  DOUBLE PRECISION, DIMENSION(NTNS,NTNS) :: expDat
  TYPE(matrixT) :: Dmat

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected result
  expDat = RESHAPE( (/ &
                    94230.7692307692d0, 40384.6153846154d0, 0.d0, &
                    40384.6153846154d0, 94230.7692307692d0, 0.d0, &
                    0.d0, 0.d0, 26923.0769230769d0 &
                    /), SHAPE(expDat) )

  !> get constitutive matrix
  CALL dmatrix(emod,nu, Dmat)
  CALL assertEquals(expDat, Dmat%dat, NTNS,NTNS, toler)

  !> deallocate constitutive matrix
  CALL dm_clean(Dmat)

END SUBROUTINE test_constitutive_VAL

END MODULE constitutive_test
