! ------------------------------------------------------------------------
!>  \brief Module defining Dense Matrix data type
! ------------------------------------------------------------------------
MODULE dense_matrix_def
USE log_message_control !> Print log/error messages
USE log_messages        !> Log/error codes
USE vector_def          !> Vector data type

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported data types
PUBLIC :: matrixT

!> Exported interfaces
PUBLIC :: dm_init, dm_clean, &
          dm_numRows, dm_numCols, &
          dm_get, dm_set, &
          dm_add, OPERATOR (+), &
          dm_scalMul, dm_vecMul, dm_matMul, OPERATOR (*), &
          dm_transpose



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************


!> Sender code for DeNSe MATrix module
INTEGER, PARAMETER :: sdr = DNSMAT



! ************************************************************************
! DATA TYPES
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Dense Matrix ADT structure
!!
!!  \param dat  Array to store matrix data
! ------------------------------------------------------------------------
TYPE matrixT
  DOUBLE PRECISION, ALLOCATABLE :: dat(:,:)
END TYPE matrixT



! ************************************************************************
! INTERFACES
! ************************************************************************


!>  \brief Interface to constructor
INTERFACE dm_init
  MODULE PROCEDURE dm_init_
  MODULE PROCEDURE dm_init_exc_
END INTERFACE dm_init

!>  \brief Interface to destructor
INTERFACE dm_clean
  MODULE PROCEDURE dm_clean_
END INTERFACE dm_clean

!>  \brief Interface to number of rows
INTERFACE dm_numRows
  MODULE PROCEDURE dm_num_rows_
END INTERFACE dm_numRows

!>  \brief Interface to number of columns
INTERFACE dm_numCols
  MODULE PROCEDURE dm_num_cols_
END INTERFACE dm_numCols

!>  \brief Interface to getter for individual entries in matrix
INTERFACE dm_get
  MODULE PROCEDURE dm_get_
  MODULE PROCEDURE dm_get_exc_
END INTERFACE dm_get

!>  \brief Interface to setter for individual entries in matrix
INTERFACE dm_set
  MODULE PROCEDURE dm_set_
  MODULE PROCEDURE dm_set_exc_
END INTERFACE dm_set

!>  \brief Interface for addition of two matrixT
INTERFACE dm_add
  MODULE PROCEDURE dm_add_
  MODULE PROCEDURE dm_add_exc_
END INTERFACE dm_add

!>  \brief Operator overload for addition
INTERFACE OPERATOR (+)
  MODULE PROCEDURE dm_add_
END INTERFACE

!>  \brief Interface for scalar multiplication
INTERFACE dm_scalMul
  MODULE PROCEDURE dm_scal_mul_mat_scal_
END INTERFACE dm_scalMul

!>  \brief Interface for matrix-vector multiplication
INTERFACE dm_vecMul
  MODULE PROCEDURE dm_vec_mul_
  MODULE PROCEDURE dm_vec_mul_exc_
END INTERFACE dm_vecMul

!>  \brief Interface for matrix-matrix multiplication
INTERFACE dm_matMul
  MODULE PROCEDURE dm_mat_mul_
  MODULE PROCEDURE dm_mat_mul_exc_
END INTERFACE dm_matMul

!>  \brief Operator overload for multiplication
!!  (includes scalar, matrix-vector, and matrix-matrix multiplication)
INTERFACE OPERATOR (*)
  MODULE PROCEDURE dm_scal_mul_mat_scal_
  MODULE PROCEDURE dm_scal_mul_scal_mat_
  MODULE PROCEDURE dm_vec_mul_
  MODULE PROCEDURE dm_mat_mul_
END INTERFACE

!>  \brief Interface for matrix transpose
INTERFACE dm_transpose
  MODULE PROCEDURE dm_transpose_
END INTERFACE dm_transpose



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Constructor for matrixT (non-exception checking)
!!
!!  \param self Reference to the dense matrix to be initialized
!!  \param m    Number of rows in the matrix
!!  \param n    Number of columns in the matrix
!!
!!  This routine initializes the matrixT object referenced by self.
!!  The data structures contained in self are allocated and initial values
!!  are set to zero.
! ------------------------------------------------------------------------
SUBROUTINE dm_init_ (self, m,n)
  TYPE(matrixT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: m,n
  INTEGER :: i,j    !> loop variables

  !> only reallocate if new dimensions do not match existing dimensions
  IF (dm_numRows(self).NE.m .OR. dm_numCols(self).NE.n) THEN

    !> reset object if it was previously initialized
    CALL dm_clean(self)

    !> allocate new data structure
    ALLOCATE(self%dat(m,n))

  END IF

  !> ensure data is initialized to zero
  DO j = 1,n
    DO i = 1,m
      self%dat(i,j) = 0.d0
    END DO
  END DO

END SUBROUTINE dm_init_


! ------------------------------------------------------------------------
!>  \brief Constructor for matrixT (exception checking)
!!
!!  \param self Reference to the dense matrix to be initialized
!!  \param m    Number of rows in the matrix
!!  \param n    Number of columns in the matrix
!!  \param exc  Error code
!!
!!  This routine initializes the matrixT object referenced by self.
!!  The data structures contained in self are allocated and initial values
!!  are set to zero. If the specified dimensions are invalid, a SZE
!!  exception is returned. If allocation of data structures fails, an
!!  ALLOC exception is returned.
! ------------------------------------------------------------------------
SUBROUTINE dm_init_exc_ (self, m,n, exc)
  TYPE(matrixT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: m,n
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: e      !> allocation info code
  INTEGER :: i,j    !> loop variables

  !> if dimensions are invalid, raise SZE exception
  IF (m.LE.0 .OR. n.LE.0) THEN
    exc=SZE
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> only reallocate if new dimensions do not match existing dimensions
  IF (dm_numRows(self).NE.m .OR. dm_numCols(self).NE.n) THEN

    !> reset object if it was previously initialized
    CALL dm_clean(self)

    !> allocate new data structure for matrix storage
    ALLOCATE(self%dat(m,n), STAT=e)

    !> if error code is returned, raise ALLOC exception
    IF (e.NE.0) THEN
      exc=ALLOC
      CALL log_printLogMsg(exc,sdr)
      RETURN
    ELSE
      exc=OK
    END IF

  END IF

  !> ensure data is initialized to zero
  DO j = 1,n
    DO i = 1,m
      self%dat(i,j) = 0.d0
    END DO
  END DO

END SUBROUTINE dm_init_exc_


! ------------------------------------------------------------------------
!>  \brief Destructor for matrixT
!!
!!  \param self Reference to the dense matrix to be destroyed
!!
!!  This routine deallocates existing data structures.
! ------------------------------------------------------------------------
SUBROUTINE dm_clean_ (self)
  TYPE(matrixT), INTENT(INOUT) :: self

  IF (ALLOCATED(self%dat)) DEALLOCATE(self%dat)

END SUBROUTINE dm_clean_


! ------------------------------------------------------------------------
!>  \brief Getter for number of rows
!!
!!  \param  self  Reference to the dense matrix object
!!  \return m     Number of rows in the matrix
!!
!!  This routine determines the number of rows allocated to the matrix
!!  object. If the matrix is not initialized it returns 0.
! ------------------------------------------------------------------------
FUNCTION dm_num_rows_ (self) RESULT(m)
  TYPE(matrixT), INTENT(IN) :: self
  INTEGER :: m

  !> if data is initialized, return number of rows
  IF (ALLOCATED(self%dat)) THEN
    m = SIZE(self%dat,1)
  ELSE
    m = 0   !> if not initialized, there are no rows
  END IF

END FUNCTION dm_num_rows_


! ------------------------------------------------------------------------
!>  \brief Getter for number of columns
!!
!!  \param  self  Reference to the dense matrix object
!!  \return n     Number of columns in the matrix
!!
!!  This routine determines the number of columns allocated to the matrix
!!  object. If the matrix is not initialized it returns 0.
! ------------------------------------------------------------------------
FUNCTION dm_num_cols_ (self) RESULT(n)
  TYPE(matrixT), INTENT(IN) :: self
  INTEGER :: n

  !> if data is initialized, return number of columns
  IF (ALLOCATED(self%dat)) THEN
    n = SIZE(self%dat,2)
  ELSE
    n = 0   !> if not initialized, there are no columns
  END IF

END FUNCTION dm_num_cols_


! ------------------------------------------------------------------------
!>  \brief Getter for matrix entries (non-exception checking)
!!
!!  \param  self  Reference to the dense matrix object
!!  \param  i     Row index
!!  \param  j     Column index
!!  \return v     Value at location (i,j) of the matrix
!!
!!  This routine determines the value at a particular location in the
!!  matrix.
! ------------------------------------------------------------------------
FUNCTION dm_get_ (self, i,j) RESULT(v)
  TYPE(matrixT), INTENT(IN) :: self
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: v

  v = self%dat(i,j)

END FUNCTION dm_get_


! ------------------------------------------------------------------------
!>  \brief Getter for matrix entries (exception checking)
!!
!!  \param  self  Reference to the dense matrix object
!!  \param  i     Row index
!!  \param  j     Column index
!!  \param  exc   Error code
!!  \return v     Value at location (i,j) of the matrix
!!
!!  This routine determines the value at a particular location in the
!!  matrix. If the location is not inside the matrix, it returns a POSIT
!!  exception.
! ------------------------------------------------------------------------
FUNCTION dm_get_exc_ (self, i,j, exc) RESULT(v)
  TYPE(matrixT), INTENT(IN) :: self
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: v

  !> make sure desired indices are inside the matrix
  IF (i.GT.dm_numRows(self) .OR. i.LE.0 &
      .OR. j.GT.dm_numCols(self) .OR. j.LE.0) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    v = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception getter
  v = dm_get(self, i,j)

END FUNCTION dm_get_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for matrix entries (non-exception checking)
!!
!!  \param self Reference to the dense matrix object
!!  \param i    Row index
!!  \param j    Column index
!!  \param v    Value to be stored at location (i,j) of the matrix
!!
!!  This routine sets the value at a particular location in the
!!  matrix.
! ------------------------------------------------------------------------
SUBROUTINE dm_set_ (self, i,j,v)
  TYPE(matrixT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: v

  self%dat(i,j) = v

END SUBROUTINE dm_set_


! ------------------------------------------------------------------------
!>  \brief Setter for matrix entries (exception checking)
!!
!!  \param self Reference to the dense matrix object
!!  \param i    Row index
!!  \param j    Column index
!!  \param v    Value to be stored at location (i,j) of the matrix
!!  \param exc  Error code
!!
!!  This routine sets the value at a particular location in the
!!  matrix. If the location is not inside the matrix, it returns a POSIT
!!  exception.
! ------------------------------------------------------------------------
SUBROUTINE dm_set_exc_ (self, i,j,v, exc)
  TYPE(matrixT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: v
  INTEGER, INTENT(OUT) :: exc

  !> make sure desired indices are inside the matrix
  IF (i.GT.dm_numRows(self) .OR. i.LE.0 &
      .OR. j.GT.dm_numCols(self) .OR. j.LE.0) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception setter
  CALL dm_set(self, i,j,v)

END SUBROUTINE dm_set_exc_


! ------------------------------------------------------------------------
!>  \brief Add two dense matrix objects (non-exception checking)
!!
!!  \param  self  Reference to the first (left-hand) dense matrix object
!!  \param  other Reference to the second (right-hand) dense matrix object
!!  \return new   Reference to the resulting banded dense object
!!
!!  This routine adds two dense matrix objects.
! ------------------------------------------------------------------------
FUNCTION dm_add_ (self,other) RESULT(new)
  TYPE(matrixT), INTENT(IN) :: self,other
  TYPE(matrixT) :: new

  !> initialize solution matrix
  CALL dm_init(new, dm_numRows(self),dm_numCols(self))

  !> Add the matrices (using array ops for efficiency)
  new%dat = self%dat + other%dat

END FUNCTION dm_add_


! ------------------------------------------------------------------------
!>  \brief Add two dense matrix objects (exception checking)
!!
!!  \param  self  Reference to the first (left-hand) dense matrix object
!!  \param  other Reference to the second (right-hand) dense matrix object
!!  \param  exc   Error code
!!  \return new   Reference to the resulting dense matrix object
!!
!!  This routine adds two dense matrix objects. If the dimensions of the
!!  two input matrices do not match, it returns a DIMEN exception.
! ------------------------------------------------------------------------
FUNCTION dm_add_exc_ (self,other, exc) RESULT(new)
  TYPE(matrixT), INTENT(IN) :: self,other
  INTEGER, INTENT(OUT) :: exc
  TYPE(matrixT) :: new

  !> make sure that the two matrices have the same number of rows
  IF(dm_numRows(self).NE.dm_numRows(other) &
          .OR. dm_numCols(self).NE.dm_numCols(other)) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  new = dm_add(self,other)

END FUNCTION dm_add_exc_


! ------------------------------------------------------------------------
!>  \brief Scalar multiplication (self*k version)
!!
!!  \param  self  Reference to the dense matrix object
!!  \param  k     Scalar multiplication factor
!!  \return new   Reference to the resulting dense matrix object
!!
!!  This routine multiplies a dense matrix by a scalar factor.
! ------------------------------------------------------------------------
FUNCTION dm_scal_mul_mat_scal_ (self, k) RESULT(new)
  TYPE(matrixT), INTENT(IN) :: self
  DOUBLE PRECISION, INTENT(IN) :: k
  TYPE(matrixT) :: new

  !> initialize solution matrix
  CALL dm_init(new, dm_numRows(self),dm_numCols(self))

  !> perform multiplication
  new%dat = k * self%dat

END FUNCTION dm_scal_mul_mat_scal_


! ------------------------------------------------------------------------
!>  \brief Scalar multiplication (k*self version)
!!
!!  \param  self  Reference to the dense matrix object
!!  \param  k     Scalar multiplication factor
!!  \return new   Reference to the resulting dense matrix object
!!
!!  This routine multiplies a dense matrix by a scalar factor. The
!!  overload is necessary for the OPERATOR (*) overload. This version
!!  simply calls the other version for better maintainability.
! ------------------------------------------------------------------------
FUNCTION dm_scal_mul_scal_mat_ (k, self) RESULT(new)
  DOUBLE PRECISION, INTENT(IN) :: k
  TYPE(matrixT), INTENT(IN) :: self
  TYPE(matrixT) :: new

  new = dm_scalMul(self,k)

END FUNCTION dm_scal_mul_scal_mat_


! ------------------------------------------------------------------------
!>  \brief Post-multiplication of a dense matrix by a vector (non-exception checking)
!!
!!  \param  self  Reference to the first (left-hand) dense matrix object
!!  \param  other Reference to the vector (right-hand) object
!!  \return new   Reference to the resulting vector object
!!
!!  This routine post-multiplies a dense matrix by a vector. That is, it
!!  performs self*other, not other*self.
! ------------------------------------------------------------------------
FUNCTION dm_vec_mul_ (self,other) RESULT(new)
  TYPE(matrixT), INTENT(IN) :: self
  TYPE(vectorT), INTENT(IN) :: other
  TYPE(vectorT) :: new

  !> initialize the solution vector
  CALL vec_init(new, dm_numRows(self))

  !> perform multiplication (using Fortran built-in for efficiency)
  new%dat = MATMUL(self%dat,other%dat)

END FUNCTION dm_vec_mul_


! ------------------------------------------------------------------------
!>  \brief Post-multiplication of a dense matrix by a vector (exception checking)
!!
!!  \param  self  Reference to the first (left-hand) dense matrix object
!!  \param  other Reference to the vector (right-hand) object
!!  \param  exc   Error code
!!  \return new   Reference to the resulting vector object
!!
!!  This routine post-multiplies a dense matrix by a vector. That is, it
!!  performs self*other, not other*self. If the dimensions of the two
!!  objects are not compatible (number of columns in self does not equal
!!  number of rows in other), it returns a DIMEN exception.
! ------------------------------------------------------------------------
FUNCTION dm_vec_mul_exc_ (self,other, exc) RESULT(new)
  TYPE(matrixT), INTENT(IN) :: self
  TYPE(vectorT), INTENT(IN) :: other
  INTEGER, INTENT(OUT) :: exc
  TYPE(vectorT) :: new

  !> make sure dimensions are compatible
  IF (dm_numCols(self).NE.vec_length(other)) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  new = dm_vecMul(self,other)

END FUNCTION dm_vec_mul_exc_


! ------------------------------------------------------------------------
!>  \brief Multiplication of two dense matrices (non-exception checking)
!!
!!  \param  self  Reference to the first (left-hand) dense matrix object
!!  \param  other Reference to the second (right-hand) dense matrix object
!!  \return new   Reference to the resulting matrix object
!!
!!  This routine multiplies two dense matrices.
! ------------------------------------------------------------------------
FUNCTION dm_mat_mul_ (self,other) RESULT(new)
  TYPE(matrixT), INTENT(IN) :: self, other
  TYPE(matrixT) :: new
  INTEGER :: m,n

  !> initialize solution matrix
  CALL dm_init(new, dm_numRows(self),dm_numCols(other))

  !> perform multiplication (using Fortran built-in for efficiency)
  new%dat = MATMUL(self%dat,other%dat)

END FUNCTION dm_mat_mul_


! ------------------------------------------------------------------------
!>  \brief Multiplication of two dense matrices (non-exception checking)
!!
!!  \param  self  Reference to the first (left-hand) dense matrix object
!!  \param  other Reference to the second (right-hand) dense matrix object
!!  \param  exc   Error code
!!  \return new   Reference to the resulting matrix object
!!
!!  This routine multiplies two dense matrices. If the dimensions of the
!!  two objects are not compatible (number of columns in self does not
!!  equal number of rows in other), it returns a DIMEN exception.
! ------------------------------------------------------------------------
FUNCTION dm_mat_mul_exc_ (self,other, exc) RESULT(new)
  TYPE(matrixT), INTENT(IN) :: self, other
  INTEGER, INTENT(OUT) :: exc
  TYPE(matrixT) :: new

  !> make sure dimensions are compatible
  IF (dm_numCols(self).NE.dm_numRows(other)) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  new = dm_matMul(self,other)

END FUNCTION dm_mat_mul_exc_


! ------------------------------------------------------------------------
!>  \brief Transpose of a matrix
!!
!!  \param  self  Reference to the input dense matrix object
!!  \return new   Reference to the transposed matrix object
! ------------------------------------------------------------------------
FUNCTION dm_transpose_ (self) RESULT (new)
  TYPE(matrixT), INTENT(IN) :: self
  TYPE(matrixT) :: new

  !> initialize result matrix
  CALL dm_init(new, dm_numCols(self),dm_numRows(self))

  !> take the transpose (using Fortran built-in for efficiency)
  new%dat = TRANSPOSE(self%dat)

END FUNCTION dm_transpose_

END MODULE dense_matrix_def



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************

! none
