! ------------------------------------------------------------------------
!>  \brief Log Message Control module
! ------------------------------------------------------------------------
MODULE log_message_control
USE system_constants    !> Global system constants (for max string length)
USE log_messages        !> Log/error message and sender codes

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************

!> Exported access programs
PUBLIC :: log_setFileName, log_getFileName, &
          log_initLogFile, log_closeLogFile, &
          log_printLogMsg



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************


INTEGER, PARAMETER :: logFile = 10          !> log file unit
CHARACTER (LEN=MAXLEN), SAVE :: logName     !> log file name



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Setter for log filename
!!
!!  \param fname  String representing filename for problem (without extension)
!!
!!  This routine takes the general filename for the problem and appends
!!  the .log extension to it for the log file.
! ------------------------------------------------------------------------
SUBROUTINE log_setFileName (fname)
  CHARACTER (LEN=*), INTENT(IN) :: fname

  logName = TRIM(fname)//'.log'

END SUBROUTINE log_setFileName


! ------------------------------------------------------------------------
!>  \brief Getter for log filename
!!
!!  \return fname  String representing filename for problem
! ------------------------------------------------------------------------
FUNCTION log_getFileName () RESULT(fname)
  CHARACTER (LEN=MAXLEN) :: fname

  fname = TRIM(logName)

END FUNCTION log_getFileName


! ------------------------------------------------------------------------
!>  \brief Initializer for log file
!!
!!  This routine initializes the file for printing log messages. It first
!!  checks if a log file with the same name as that set in the module
!!  exists (and deletes it, if present).
! ------------------------------------------------------------------------
SUBROUTINE log_initLogFile ()
  LOGICAL :: fileExists   !> boolean for checking file existence

  !> if file already exists, delete it
  INQUIRE(FILE=logName, EXIST=fileExists)
  IF (fileExists) THEN
    OPEN(UNIT=logFile, FILE=logName, STATUS='OLD')
    CLOSE(UNIT=logFile, STATUS='DELETE')
  END IF

  !> open the new file
  OPEN(UNIT=logFile, FILE=logName)

END SUBROUTINE log_initLogFile


! ------------------------------------------------------------------------
!>  \brief Finalizer for log file
!!
!!  This routine first checks if the log file is open, and if so closes
!!  it.
! ------------------------------------------------------------------------
SUBROUTINE log_closeLogFile ()
  LOGICAL :: fileOpened   !> boolean for checking if file is open

  !> close the log file if it is open
  INQUIRE(UNIT=logFile, OPENED=fileOpened)
  IF (fileOpened) THEN
    CLOSE(UNIT=logFile)
  END IF

END SUBROUTINE log_closeLogFile


! ------------------------------------------------------------------------
!>  \brief Print log messages to the log file
!!
!!  \param msg  Message code (messageT)
!!  \param sdr  Sender code (senderT)
!!
!!  This routine takes the message code and sender code and turns them
!!  into their associated human readable strings, which it then prints
!!  to the log file. It makes sure that the log file has been initialized
!!  and, if not, does not try to print.
! ------------------------------------------------------------------------
SUBROUTINE log_printLogMsg (msg, sdr)
  INTEGER, INTENT(IN) :: msg, sdr
  LOGICAL :: fileOpened   !> boolean for making sure log file is open

  !> check that log file is open
  INQUIRE(UNIT=logFile, OPENED=fileOpened)

  !> if so, print the messages
  IF (fileOpened) THEN
    WRITE(UNIT=logFile, FMT='(A)') 'Message: '//msg_getMsg(msg)
    WRITE(UNIT=logFile, FMT='(A)') 'Sender: '//msg_getSdr(sdr)
    WRITE(UNIT=logFile, FMT=*)
  END IF

END SUBROUTINE log_printLogMsg


END MODULE log_message_control
