! ------------------------------------------------------------------------
!>  \brief Module for testing Log Message Control module
! ------------------------------------------------------------------------
MODULE log_message_control_test
USE fruit                 !> Unit testing framework
USE log_message_control   !> Print log/error messages
USE log_messages          !> Log/error message and sender codes

IMPLICIT NONE


CONTAINS


! ------------------------------------------------------------------------
!>  \test Test for setting of log message filename
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file
!!  \param expected   Expected filename with extension
!!
!!  This test checks that the filename for the log file is parsed and
!!  appended with the correct extension.
! ------------------------------------------------------------------------
SUBROUTINE test_log_setFileName
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_log_setFileName'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  CHARACTER (LEN=*), PARAMETER :: expected = 'testName.log'

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set the log filename
  CALL log_setFileName(testName)

  !> check that correct filename is set
  CALL assertEquals(expected, log_getFileName() )

END SUBROUTINE test_log_setFileName


! ------------------------------------------------------------------------
!>  \test Test for log file initialization
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file
!!  \param fileExists Boolean for checking existence of log file
!!
!!  This test checks that the log file is correctly created.
! ------------------------------------------------------------------------
SUBROUTINE test_log_initLogFile
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_log_initLogFile'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  LOGICAL :: fileExists

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set file name and check if it exists
  CALL log_setFileName(testName)
  INQUIRE(FILE=log_getFileName(), EXIST=fileExists)

  !> if file already exists, delete it
  IF (fileExists) THEN
    OPEN(UNIT=1, FILE=log_getFileName(), STATUS='OLD')
    CLOSE(UNIT=1, STATUS='DELETE')
  END IF

  !> initialize a new file and make sure it exists
  CALL log_initLogFile()
  INQUIRE(FILE=log_getFileName(), EXIST=fileExists)
  CALL assertEquals(.TRUE., fileExists)

END SUBROUTINE test_log_initLogFile


! ------------------------------------------------------------------------
!>  \test Test for log file finalization
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file
!!  \param fileOpened Boolean for checking whether file is open
!!
!!  This test checks that the log file is correctly closed.
! ------------------------------------------------------------------------
SUBROUTINE test_log_closeLogFile
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_log_closeLogFile'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  LOGICAL :: fileOpened

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set file name and initialize
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> close log file and make sure file is not associated with a unit
  CALL log_closeLogFile()
  INQUIRE(FILE=log_getFileName(), OPENED=fileOpened)
  CALL assertEquals(.FALSE., fileOpened)

END SUBROUTINE test_log_closeLogFile


! ------------------------------------------------------------------------
!>  \test Test of log message printing
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file
!!  \param fileOpened Boolean for checking whether file is open
!!
!!  This test checks for correct printing of a log message to the log
!!  file.
! ------------------------------------------------------------------------
SUBROUTINE test_log_printLogMsg
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_log_printLogMsg'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  CHARACTER (LEN=*), PARAMETER :: expMsg = 'Message: Failed to allocate memory'
  CHARACTER (LEN=*), PARAMETER :: expSdr = 'Sender: Body Force Reader'
  CHARACTER (LEN=MAXLEN) :: actMsg, actSdr

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set file name and initialize
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> write log message
  CALL log_printLogMsg(ALLOC, BFCRDR)

  !> close log file
  CALL log_closeLogFile()

  !> open log file and read for test
  OPEN(UNIT=1, FILE=log_getFileName())
  READ(UNIT=1, FMT='(A)', ERR=1, END=1) actMsg
  READ(UNIT=1, FMT='(A)', ERR=1, END=1) actSdr
1 CLOSE(UNIT=1)   !> label must be used here to avoid file reading errors

  !> check that correct messages were written
  CALL assertEquals(expMsg, actMsg)
  CALL assertEquals(expSdr, actSdr)

END SUBROUTINE test_log_printLogMsg

END MODULE log_message_control_test
