! ------------------------------------------------------------------------
!>  \brief Module for Material Models
! ------------------------------------------------------------------------
MODULE material_model
USE dense_matrix_def    !> Dense Matrix data type
USE vector_def          !> Vector data type
USE constitutive        !> Constitutive Matrix module

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported interfaces
PUBLIC :: linearElastic



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************

! none


! ************************************************************************
! DATA TYPES
! ************************************************************************

! none



! ************************************************************************
! INTERFACES
! ************************************************************************


!> \brief Interface to linear elastic material model
INTERFACE linearElastic
  MODULE PROCEDURE linear_elastic_2d_plane_strain_
END INTERFACE linearElastic



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Linear elastic material model (2-d, plane strain)
!!
!!  \param  emod  Elastic modulus
!!  \param  nu    Poisson's ratio
!!  \param  deps  Strain increment
!!  \param  dsig  Stress increment
!!
!!  This routine computes the incremental stress given the incremental
!!  strain for a linear elastic material.
! ------------------------------------------------------------------------
SUBROUTINE linear_elastic_2d_plane_strain_ (emod,nu, deps, dsig)
  DOUBLE PRECISION, INTENT(IN) :: emod,nu
  TYPE(vectorT), INTENT(IN) :: deps
  TYPE(vectorT), INTENT(INOUT) :: dsig
  TYPE(matrixT) :: Dmat   !> constitutive matrix

  !> get constitutive matrix
  CALL dmatrix(emod,nu, Dmat)

  CALL vec_clean(dsig)    !> ensure dsig is deallocated (avoid memory leak)
  dsig = Dmat*deps        !> compute stress increment

  !> deallocate constitutive matrix
  CALL dm_clean(Dmat)

END SUBROUTINE linear_elastic_2d_plane_strain_



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************

! none

END MODULE material_model
