! ------------------------------------------------------------------------
!>  \brief Module for PDE Solver Control
! ------------------------------------------------------------------------
MODULE pde_solver_control
USE system_constants              !> global system constants
USE pde_solver_constants          !> constants for PDE solver
USE log_message_control           !> print log/error messages
USE log_messages                  !> log/error message and sender codes
USE band_sym_matrix_def           !> banded symmetric matrix ADT
USE dense_matrix_def              !> dense matrix ADT
USE vector_def                    !> vector ADT
USE field_data                    !> field data module
USE boundary_data                 !> boundary data module
USE material_data                 !> material property data module
USE body_element_integration      !> body element integration (mass and stiff)
USE traction_element_integration  !> traction element load vectors

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported state variables
PUBLIC :: hbw, nnod, nel, nelb, ndof,                 &
          mass, modMass, damp, stiff,                 &
          initStress, initStrain, body, trac, load,   &
          prevDisp, incDisp, newDisp,                 &
          prevVel, incVel, newVel,                    &
          prevAcc, newAcc,                            &
          prevStress, incStress, newStress,           &
          prevStrain, incStrain, newStrain


!> Exported interfaces
PUBLIC :: pde_init, pde_clean, &
          pde_buildMassMatrix, &
          pde_buildStiffMatrix!, &
!          pde_buildDampMatrix, &
!          pde_buildModMassMatrix, &
!          pde_buildLoadVector, &
!          pde_initAcc, &
!          pde_incAcc, pde_incVel, pde_incDisp, &
!          pde_incStress, pde_incStrain, &
!          pde_updateAcc, pde_updateVel, pde_updateDisp, &
!          pde_updateStress, pde_updateStrain



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************

! none



! ************************************************************************
! DATA TYPES
! ************************************************************************

! none



! ************************************************************************
! STATE VARIABLES
! ************************************************************************

INTEGER, SAVE               :: hbw            !> half bandwidth of system matrices
INTEGER, SAVE               :: nnod           !> total number of nodes
INTEGER, SAVE               :: nel            !> total number of body elements
INTEGER, SAVE               :: nelb           !> total number of traction elements
INTEGER, SAVE               :: ndof           !> total number of system degrees of freedom
TYPE(bandSymMatrixT), SAVE  :: mass           !> mass matrix
TYPE(bandSymMatrixT), SAVE  :: modMass        !> modified mass matrix (for time-stepping)
TYPE(bandSymMatrixT), SAVE  :: stiff          !> stiffness matrix
TYPE(bandSymMatrixT), SAVE  :: damp           !> damping matrix
TYPE(vectorT), SAVE         :: initStress     !> load vector due to initial stress
TYPE(vectorT), SAVE         :: initStrain     !> load vector due to initial strain
TYPE(vectorT), SAVE         :: body           !> load vector due to body forces
TYPE(vectorT), SAVE         :: trac           !> load vector due to surface tractions
TYPE(vectorT), SAVE         :: load           !> total load vector
TYPE(vectorT), SAVE         :: prevDisp       !> displacement vector on previous time step
TYPE(vectorT), SAVE         :: incDisp        !> change in displacement on current time step
TYPE(vectorT), SAVE         :: newDisp        !> displacement vector on current time step
TYPE(vectorT), SAVE         :: prevVel        !> velocity vector on previous time step
TYPE(vectorT), SAVE         :: incVel         !> change in velocity on current time step
TYPE(vectorT), SAVE         :: newVel         !> velocity vector on current time step
TYPE(vectorT), SAVE         :: prevAcc        !> acceleration vector on previous time step
TYPE(vectorT), SAVE         :: newAcc         !> acceleration vector on current time step
TYPE(vectorT), SAVE         :: prevStress     !> element stresses on previous time step
TYPE(vectorT), SAVE         :: incStress      !> change in element stresses on current time step
TYPE(vectorT), SAVE         :: newStress      !> element stresses on current time step
TYPE(vectorT), SAVE         :: prevStrain     !> element strains on previous time step
TYPE(vectorT), SAVE         :: incStrain      !> change in element strains on current time step
TYPE(vectorT), SAVE         :: newStrain      !> element strains on current time step



! ************************************************************************
! INTERFACES
! ************************************************************************

!> \brief Interface to PDE solver initializer
INTERFACE pde_init
  MODULE PROCEDURE pde_init_
END INTERFACE pde_init

!> \brief Interface to PDE solver destructor
INTERFACE pde_clean
  MODULE PROCEDURE pde_clean_
END INTERFACE pde_clean

!> \brief Interface to mass matrix constructor
INTERFACE pde_buildMassMatrix
  MODULE PROCEDURE pde_build_mass_matrix_
END INTERFACE pde_buildMassMatrix

!> \brief Interface to stiffness matrix constructor
INTERFACE pde_buildStiffMatrix
  MODULE PROCEDURE pde_build_stiff_matrix_
END INTERFACE pde_buildStiffMatrix

!> \brief Interface to damping matrix constructor
!INTERFACE pde_buildDampMatrix
!  MODULE PROCEDURE pde_build_damp_matrix_
!END INTERFACE pde_buildDampMatrix

!> \brief Interface to modified mass matrix constructor
!INTERFACE pde_buildModMassMatrix
!  MODULE PROCEDURE pde_build_mod_mass_matrix_
!END INTERFACE pde_buildModMassMatrix

!> \brief Interface to load vector constructor
!INTERFACE pde_buildLoadVector
!  MODULE PROCEDURE pde_build_load_vector_
!END INTERFACE pde_buildLoadVector

!> \brief Interface to acceleration vector initializer
!INTERFACE pde_initAcc
!  MODULE PROCEDURE pde_init_acc_
!END INTERFACE pde_initAcc

!> \brief Interface to acceleration vector incrementer
!INTERFACE pde_incAcc
!  MODULE PROCEDURE pde_inc_acc_
!END INTERFACE pde_incAcc

!> \brief Interface to velocity vector incrementer
!INTERFACE pde_incVel
!  MODULE PROCEDURE pde_inc_vel_
!END INTERFACE pde_incVel

!> \brief Interface to displacement vector incrementer
!INTERFACE pde_incDisp
!  MODULE PROCEDURE pde_inc_disp_
!END INTERFACE pde_incDisp

!> \brief Interface to stress incrementer
!INTERFACE pde_incStress
!  MODULE PROCEDURE pde_inc_stress_
!END INTERFACE pde_incStress

!> \brief Interface to strain incrementer
!INTERFACE pde_incStrain
!  MODULE PROCEDURE pde_inc_strain_
!END INTERFACE pde_incStrain

!> \brief Interface to acceleration vector updater
!INTERFACE pde_updateAcc
!  MODULE PROCEDURE pde_update_acc_
!END INTERFACE pde_updateAcc

!> \brief Interface to velocity vector updater
!INTERFACE pde_updateVel
!  MODULE PROCEDURE pde_update_vel_
!END INTERFACE pde_updateVel

!> \brief Interface to displacement vector updater
!INTERFACE pde_updateDisp
!  MODULE PROCEDURE pde_update_disp_
!END INTERFACE pde_updateDisp

!> \brief Interface to stress updater
!INTERFACE pde_updateStress
!  MODULE PROCEDURE pde_update_stress_
!END INTERFACE pde_updateStress

!> \brief Interface to strain updater
!INTERFACE pde_updateStrain
!  MODULE PROCEDURE pde_update_strain_
!END INTERFACE pde_updateStrain



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Initializer for PDE solver
!!
!!  \param  exc For returning exception codes. (Optional).
!!
!!  This routine sets up the solution space for the problem by allocating
!!  memory for the following state variables:
!!
!!    mass, stiff, initStress, initStrain, body, trac, load,
!!    prevDisp, incDisp, newDisp,
!!    prevVel, incVel, newVel,
!!    prevAcc, newAcc,
!!    prevStress, incStress, newStress,
!!    prevStrain, incStrain, newStrain
!!
!!  Note that modMass and damp are not allocated here since
!!  they will be computed by combining other matrices later.
!!
!!  This routine also initializes the values stored in:
!!
!!    prevDisp, prevVel, prevStress, prevStrain
!!
!!  These represent the initial conditions of the system.
! ------------------------------------------------------------------------
SUBROUTINE pde_init_ (exc)
  INTEGER, INTENT(OUT), OPTIONAL :: exc
  INTEGER :: exc_tmp    !> for storing exceptions
  INTEGER :: nstress    !> total number of stress/strain entries
  INTEGER :: i,j        !> loop variables
  INTEGER :: dof        !> degree of freedom number

  !> determine system size properties
  nnod  = fld_numNode()
  nel   = fld_numElem()
  nelb  = bnd_numBoundElem()
  ndof  = fld_numDof()
  hbw   = compute_hbw()

  ! ----------------------------
  !> allocate system matrices
  ! ----------------------------

  !> mass matrix
  CALL bsm_init(mass, hbw,ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> stiffness matrix
  CALL bsm_init(stiff, hbw,ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF


  ! ----------------------------
  !> allocate system vectors
  ! ----------------------------

  !> initial stress load vector
  CALL vec_init(initStress, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> initial strain load vector
  CALL vec_init(initStrain, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> body force load vector
  CALL vec_init(body, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> traction load vector
  CALL vec_init(trac, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> total load vector
  CALL vec_init(load, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> previous displacement
  CALL vec_init(prevDisp, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> incremental displacement
  CALL vec_init(incDisp, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> current displacement
  CALL vec_init(newDisp, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> previous velocity
  CALL vec_init(prevVel, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> incremental velocity
  CALL vec_init(incVel, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> current velocity
  CALL vec_init(newVel, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> previous acceleration
  CALL vec_init(prevAcc, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> current acceleration
  CALL vec_init(newAcc, ndof, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF


  ! ----------------------------------
  !> allocate stress/strain vectors
  ! ----------------------------------

  nstress = nel*NTNS

  !> previous stress
  CALL vec_init(prevStress, nstress, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> incremental stress
  CALL vec_init(incStress, nstress, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> current stress
  CALL vec_init(newStress, nstress, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> previous strain
  CALL vec_init(prevStrain, nstress, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> incremental strain
  CALL vec_init(incStrain, nstress, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF

  !> current strain
  CALL vec_init(newStrain, nstress, exc_tmp)
  IF (exc_tmp.NE.OK) THEN
    IF (PRESENT(exc)) exc=exc_tmp
    RETURN
  END IF


  ! ----------------------------
  !> set up initial conditions
  ! ----------------------------

  !> displacement and velocity
  DO i = 1,nnod
    DO j = 1,NDIM

      dof = fld_getDof(i,j, exc_tmp)
      IF (exc_tmp.NE.OK) THEN
        IF (PRESENT(exc)) exc=exc_tmp
        RETURN
      END IF

      IF (dof.NE.0) THEN

        CALL vec_set(prevDisp, dof, fld_getDisp(i,j), exc_tmp)
        IF (exc_tmp.NE.OK) THEN
          IF (PRESENT(exc)) exc=exc_tmp
          RETURN
        END IF

        CALL vec_set(prevVel, dof, fld_getVel(i,j), exc_tmp)
        IF (exc_tmp.NE.OK) THEN
          IF (PRESENT(exc)) exc=exc_tmp
          RETURN
        END IF

      END IF

    END DO
  END DO

  !> stress and strain
  DO i = 1,nel

    dof = (i-1)*NTNS

    DO j = 1,NTNS

      CALL vec_set(prevStress, dof+j, fld_getStressElem(i,j), exc_tmp)
      IF (exc_tmp.NE.OK) THEN
        IF (PRESENT(exc)) exc=exc_tmp
        RETURN
      END IF

      CALL vec_set(prevStrain, dof+j, fld_getStrainElem(i,j), exc_tmp)
      IF (exc_tmp.NE.OK) THEN
        IF (PRESENT(exc)) exc=exc_tmp
        RETURN
      END IF

    END DO

  END DO


  !> if no exceptions were raised, everything is A-OK
  IF (PRESENT(exc)) exc=OK

END SUBROUTINE pde_init_


! ------------------------------------------------------------------------
!>  \brief Destructor for PDE solver
!!
!!  This routine ensures that all state variables that have been
!!  dynamically allocated are deallocated. It also resets system size
!!  parameters.
! ------------------------------------------------------------------------
SUBROUTINE pde_clean_ ()

  !> size parameters
  nnod  = 0
  nel   = 0
  nelb  = 0
  ndof  = 0
  hbw   = 0

  !> matrices
  CALL bsm_clean(mass)
  CALL bsm_clean(modMass)
  CALL bsm_clean(stiff)
  CALL bsm_clean(damp)

  !> vectors
  CALL vec_clean(initStress)
  CALL vec_clean(initStrain)
  CALL vec_clean(body)
  CALL vec_clean(trac)
  CALL vec_clean(load)
  CALL vec_clean(prevDisp)
  CALL vec_clean(incDisp)
  CALL vec_clean(newDisp)
  CALL vec_clean(prevVel)
  CALL vec_clean(incVel)
  CALL vec_clean(newVel)
  CALL vec_clean(prevAcc)
  CALL vec_clean(newAcc)
  CALL vec_clean(prevStress)
  CALL vec_clean(incStress)
  CALL vec_clean(newStress)
  CALL vec_clean(prevStrain)
  CALL vec_clean(incStrain)
  CALL vec_clean(newStrain)

END SUBROUTINE pde_clean_


! ------------------------------------------------------------------------
!>  \brief  Build the global mass matrix
!!
!!  \param  exc   Error code
!!
!!  This routine constructs the global (consistent) mass matrix by
!!  summing element mass matrices taking connectivity into account. This
!!  routine assumes that pde_init() has already been called.
! ------------------------------------------------------------------------
SUBROUTINE pde_build_mass_matrix_ (exc)
  INTEGER, INTENT(OUT), OPTIONAL :: exc
  TYPE(matrixT) :: emass      !> element mass matrix
  INTEGER :: ind(NNODEL*NDIM) !> mapping indices
  INTEGER :: i    !> loop variable
  INTEGER :: exc_tmp    !> for getting exception code

  !> loop through elements, adding element mass matrices
  DO i = 1,nel

    !> get element mass matrix and add it to the global mass matrix
    CALL bint_emass(i,emass)
    ind = get_index(i)
    CALL bsm_mappedAdd(mass, emass,ind, exc_tmp)

    !> check that mapped add was successful
    IF (exc_tmp.NE.OK) THEN
      IF (PRESENT(exc)) exc=exc_tmp
      RETURN
    END IF

  END DO

  !> if loop completed, everything is OK
  IF (PRESENT(exc)) exc=OK

END SUBROUTINE pde_build_mass_matrix_


! ------------------------------------------------------------------------
!>  \brief  Build the global stiffness matrix
!!
!!  \param  exc   Error code
!!
!!  This routine constructs the global stiffness matrix by summing element
!!  stiffness matrices taking connectivity into account. This
!!  routine assumes that pde_init() has already been called.
! ------------------------------------------------------------------------
SUBROUTINE pde_build_stiff_matrix_ (exc)
  INTEGER, INTENT(OUT), OPTIONAL :: exc
  TYPE(matrixT) :: estiff      !> element stiffness matrix
  INTEGER :: ind(NNODEL*NDIM) !> mapping indices
  INTEGER :: i    !> loop variable
  INTEGER :: exc_tmp    !> for getting exception code

  !> loop through elements, adding element stiffness matrices
  DO i = 1,nel

    !> get element stiffness matrix and add it to the global stiffness matrix
    CALL bint_estiff(i,estiff)
    ind = get_index(i)
    CALL bsm_mappedAdd(stiff, estiff,ind, exc_tmp)

    !> check that mapped add was successful
    IF (exc_tmp.NE.OK) THEN
      IF (PRESENT(exc)) exc=exc_tmp
      RETURN
    END IF

  END DO

  !> if loop completed, everything is OK
  IF (PRESENT(exc)) exc=OK

END SUBROUTINE pde_build_stiff_matrix_



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief  Compute the half bandwidth of the system matrices
!!
!!  \return hbw   Half bandwidth
! ------------------------------------------------------------------------
FUNCTION compute_hbw () RESULT(hbw)
  INTEGER :: hbw
  INTEGER :: i,j,k,l,m  !> loop variables
  INTEGER :: lmin, mmin !> loop limits
  INTEGER :: nel        !> number of elements
  INTEGER :: dof1,dof2  !> for calculating diff between dofs

  !> initialize half bandwidth
  hbw = 0

  !> get number of elements
  nel = fld_numElem()

  !> loop through elements
  !! (the goal is to find the maximum absolute difference between
  !!  non-zero degree of freedom numbers within a single element)
  DO i = 1,nel
    DO j = 1,NNODEL   !> loop through nodes in current element
      DO k = 1,NDIM   !> loop through coordinate (dof) directions

        !> get current dof
        dof1 = fld_getDof(fld_getConnect(i,j),k)

        !> if current dof=0, no need to compare
        IF (dof1.EQ.0)  CYCLE

        !> get minimum loop indices for next dof
        IF (k.LT.NDIM) THEN
          lmin = j    !> start from current node
          mmin = k+1  !> start from next dof of current node
        ELSE
          lmin = j+1  !> start from next node
          mmin = 1    !> start from first dof of next node
        END IF

        DO l = lmin,NNODEL
          DO m = mmin,NDIM

            !> get next dof
            dof2 = fld_getDof(fld_getConnect(i,l),m)

            !> if next dof is non-zero, update hbw
            IF (dof2.NE.0) THEN
              hbw = MAX(hbw, ABS(dof2-dof1))
            END IF

          END DO  ! m = mmin,NDIM
        END DO  ! l = lmin,NNODEL

      END DO  ! k = 1,NDIM
    END DO  ! j = 1,NNODEL
  END DO  ! i = 1,nel

  !> add one to include diagonal
  hbw = hbw+1

END FUNCTION compute_hbw


! ------------------------------------------------------------------------
!>  \brief  Determine the mapping indices for a body element
!!
!!  \param  i     Element number
!!
!!  \return ind   Mapping indices
!!
!!  Note that this function assumes that the element index is valid. It
!!  does not catch POSIT exceptions
! ------------------------------------------------------------------------
FUNCTION get_index (i) RESULT(ind)
  INTEGER, INTENT(IN) :: i
  INTEGER :: ind(NNODEL*NDIM)
  INTEGER :: j,k    !> loop variables

  DO j = 1,NNODEL
    DO k = 1,NDIM
      ind( (j-1)*NDIM+k ) = fld_getDof(fld_getConnect(i,j),k)
    END DO
  END DO

END FUNCTION


! ------------------------------------------------------------------------
!>  \brief  Determine the mapping indices for a traction element
!!
!!  \param  i     Element number
!!
!!  \return ind   Mapping indices
!!
!!  Note that this function assumes that the element index is valid. It
!!  does not catch POSIT exceptions
! ------------------------------------------------------------------------
FUNCTION get_index_trac (i) RESULT(ind)
  INTEGER, INTENT(IN) :: i
  INTEGER :: ind(NNODELB*NDIM)
  INTEGER :: j,k    !> loop variables

  DO j = 1,NNODELB
    DO k = 1,NDIM
      ind( (j-1)*NDIM+k ) = fld_getDof(bnd_getConnect(i,j),k)
    END DO
  END DO

END FUNCTION

END MODULE pde_solver_control
