! ------------------------------------------------------------------------
!>  \brief Module defining Vector data type
! ------------------------------------------------------------------------
MODULE vector_def
USE log_message_control !> Print log/error messages
USE log_messages        !> Log/error codes

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported data types
PUBLIC :: vectorT

!> Exported interfaces
PUBLIC :: vec_init, vec_clean, &
          vec_length, &
          vec_get, vec_set, &
          vec_add, OPERATOR (+), vec_mappedAdd, &
          vec_scalMul, OPERATOR (*), vec_dotProd



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************


!> Sender code for VECTOR module
INTEGER, PARAMETER :: sdr = VECTOR



! ************************************************************************
! DATA TYPES
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Vector ADT structure
!!
!!  \param dat  Array to store vector data
! ------------------------------------------------------------------------
TYPE vectorT
  DOUBLE PRECISION, ALLOCATABLE :: dat(:)
END TYPE vectorT



! ************************************************************************
! INTERFACES
! ************************************************************************


!>  \brief Interface to constructor
INTERFACE vec_init
  MODULE PROCEDURE vec_init_
  MODULE PROCEDURE vec_init_exc_
END INTERFACE vec_init

!>  \brief Interface to destructor
INTERFACE vec_clean
  MODULE PROCEDURE vec_clean_
END INTERFACE vec_clean

!>  \brief Interface to length of vector
INTERFACE vec_length
  MODULE PROCEDURE vec_length_
END INTERFACE vec_length

!>  \brief Interface to getter for individual entries in vector
INTERFACE vec_get
  MODULE PROCEDURE vec_get_
  MODULE PROCEDURE vec_get_exc_
END INTERFACE vec_get

!>  \brief Interface to setter for individual entries in vector
INTERFACE vec_set
  MODULE PROCEDURE vec_set_
  MODULE PROCEDURE vec_set_exc_
END INTERFACE vec_set

!>  \brief Interface for addition of two vectorT
INTERFACE vec_add
  MODULE PROCEDURE vec_add_
  MODULE PROCEDURE vec_add_exc_
END INTERFACE vec_add

!>  \brief Operator overload for addition
INTERFACE OPERATOR (+)
  MODULE PROCEDURE vec_add_
END INTERFACE

!>  \brief Interface for adding a vectorT to a vectorT in place with mapping indices
INTERFACE vec_mappedAdd
  MODULE PROCEDURE vec_mapped_add_
  MODULE PROCEDURE vec_mapped_add_exc_
END INTERFACE vec_mappedAdd

!>  \brief Interface for scalar multiplication
INTERFACE vec_scalMul
  MODULE PROCEDURE vec_scal_mul_vec_scal_
END INTERFACE vec_scalMul

!>  \brief Operator overload for multiplication
INTERFACE OPERATOR (*)
  MODULE PROCEDURE vec_scal_mul_vec_scal_
  MODULE PROCEDURE vec_scal_mul_scal_vec_
END INTERFACE

!>  \brief Interface to dot product of two vectorT
INTERFACE vec_dotProd
  MODULE PROCEDURE vec_dot_prod_
  MODULE PROCEDURE vec_dot_prod_exc_
END INTERFACE vec_dotProd



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Constructor for vectorT (non-exception checking)
!!
!!  \param self Reference to the vector to be initialized
!!  \param n    Number of entries in the vector
!!
!!  This routine initializes the vectorT object referenced by self.
!!  The data structures contained in self are allocated and initial values
!!  are set to zero.
! ------------------------------------------------------------------------
SUBROUTINE vec_init_ (self, n)
  TYPE(vectorT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: n
  INTEGER :: i    !> loop variable

  !> only reallocate if new dimensions do not match existing dimensions
  IF (vec_length(self).NE.n) THEN

    !> reset object if it was previously initialized
    CALL vec_clean(self)

    !> allocate new data structure
    ALLOCATE(self%dat(n))

  END IF

  !> ensure data is initialized to zero
  DO i = 1,n
    self%dat(i) = 0.d0
  END DO

END SUBROUTINE vec_init_


! ------------------------------------------------------------------------
!>  \brief Constructor for vectorT (exception checking)
!!
!!  \param self Reference to the vector to be initialized
!!  \param n    Number of entries in the vector
!!  \param exc  Error code
!!
!!  This routine initializes the vectorT object referenced by self.
!!  The data structures contained in self are allocated and initial values
!!  are set to zero. If the specified dimension is invalid, a SZE
!!  exception is returned. If allocation of data structures fails, an
!!  ALLOC exception is returned.
! ------------------------------------------------------------------------
SUBROUTINE vec_init_exc_ (self, n, exc)
  TYPE(vectorT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: n
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: e    !> allocation info code
  INTEGER :: i    !> loop variable

  !> if dimension is invalid, raise SZE exception
  IF (n.LE.0) THEN
    exc=SZE
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> only reallocate if new dimensions do not match existing dimensions
  IF (vec_length(self).NE.n) THEN

    !> reset object if it was previously initialized
    CALL vec_clean(self)

    !> allocate new data structure for vector storage
    ALLOCATE(self%dat(n), STAT=e)

    !> if error code is returned, raise ALLOC exception
    IF (e.NE.0) THEN
      exc=ALLOC
      CALL log_printLogMsg(exc,sdr)
      RETURN
    ELSE
      exc=OK
    END IF

  END IF

  !> ensure data is initialized to zero
  DO i = 1,n
    self%dat(i) = 0.d0
  END DO

END SUBROUTINE vec_init_exc_


! ------------------------------------------------------------------------
!>  \brief Destructor for vectorT
!!
!!  \param self Reference to the vector to be destroyed
!!
!!  This routine deallocates existing data structures.
! ------------------------------------------------------------------------
SUBROUTINE vec_clean_ (self)
  TYPE(vectorT), INTENT(INOUT) :: self

  IF (ALLOCATED(self%dat)) DEALLOCATE(self%dat)

END SUBROUTINE vec_clean_


! ------------------------------------------------------------------------
!>  \brief Getter for length of vector
!!
!!  \param  self  Reference to the vector object
!!  \return n     Number of entries in the vector
!!
!!  This routine determines the number of entries allocated to the vector
!!  object. If the vector is not initialized it returns 0.
! ------------------------------------------------------------------------
FUNCTION vec_length_ (self) RESULT(n)
  TYPE(vectorT), INTENT(IN) :: self
  INTEGER :: n

  !> if data is initialized, return number of entries
  IF (ALLOCATED(self%dat)) THEN
    n = SIZE(self%dat)
  ELSE
    n = 0   !> if not initialized, there are no entries
  END IF

END FUNCTION vec_length_


! ------------------------------------------------------------------------
!>  \brief Getter for vector entries (non-exception checking)
!!
!!  \param  self  Reference to the vector object
!!  \param  i     Location index
!!  \return v     Value at location (i) of the vector
!!
!!  This routine determines the value at a particular location in the
!!  vector.
! ------------------------------------------------------------------------
FUNCTION vec_get_ (self, i) RESULT(v)
  TYPE(vectorT), INTENT(IN) :: self
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION :: v

  v = self%dat(i)

END FUNCTION vec_get_


! ------------------------------------------------------------------------
!>  \brief Getter for vector entries (exception checking)
!!
!!  \param  self  Reference to the vector object
!!  \param  i     Location index
!!  \param  exc   Error code
!!  \return v     Value at location (i) of the vector
!!
!!  This routine determines the value at a particular location in the
!!  vector. If the location is not inside the vector, it returns a POSIT
!!  exception.
! ------------------------------------------------------------------------
FUNCTION vec_get_exc_ (self, i, exc) RESULT(v)
  TYPE(vectorT), INTENT(IN) :: self
  INTEGER, INTENT(IN) :: i
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: v

  !> make sure desired index are is the vector
  IF (i.GT.vec_length(self) .OR. i.LE.0) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    v = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception getter
  v = vec_get(self, i)

END FUNCTION vec_get_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for vector entries (non-exception checking)
!!
!!  \param  self  Reference to the vector object
!!  \param  i     Location index
!!  \param v      Value to be stored at location (i) of the vector
!!
!!  This routine sets the value at a particular location in the vector.
! ------------------------------------------------------------------------
SUBROUTINE vec_set_ (self, i,v)
  TYPE(vectorT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION, INTENT(IN) :: v

  self%dat(i) = v

END SUBROUTINE vec_set_


! ------------------------------------------------------------------------
!>  \brief Setter for vector entries (exception checking)
!!
!!  \param  self  Reference to the vector object
!!  \param  i     Location index
!!  \param  exc   Error code
!!  \return v     Value to be stored at location (i) of the vector
!!
!!  This routine sets the value at a particular location in the vector. If
!!  the location is not inside the vector, it returns a POSIT exception.
! ------------------------------------------------------------------------
SUBROUTINE vec_set_exc_ (self, i,v, exc)
  TYPE(vectorT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION, INTENT(IN) :: v
  INTEGER, INTENT(OUT) :: exc

  !> make sure desired index is inside the vector
  IF (i.GT.vec_length(self) .OR. i.LE.0) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception setter
  CALL vec_set(self, i,v)

END SUBROUTINE vec_set_exc_


! ------------------------------------------------------------------------
!>  \brief Add two vector objects (non-exception checking)
!!
!!  \param  self  Reference to the first (left-hand) vector object
!!  \param  other Reference to the second (right-hand) vector object
!!  \return new   Reference to the resulting vector object
!!
!!  This routine adds two vector objects.
! ------------------------------------------------------------------------
FUNCTION vec_add_ (self,other) RESULT(new)
  TYPE(vectorT), INTENT(IN) :: self,other
  TYPE(vectorT) :: new

  !> initialize solution vector
  CALL vec_init(new, vec_length(self))

  !> Add the vectors (using array ops for efficiency)
  new%dat = self%dat + other%dat

END FUNCTION vec_add_


! ------------------------------------------------------------------------
!>  \brief Add two vector objects (exception checking)
!!
!!  \param  self  Reference to the first (left-hand) vector object
!!  \param  other Reference to the second (right-hand) vector object
!!  \param  exc   Error code
!!  \return new   Reference to the resulting vector object
!!
!!  This routine adds two vector objects. If the dimensions of the
!!  two input vectors do not match, it returns a DIMEN exception.
! ------------------------------------------------------------------------
FUNCTION vec_add_exc_ (self,other, exc) RESULT(new)
  TYPE(vectorT), INTENT(IN) :: self,other
  INTEGER, INTENT(OUT) :: exc
  TYPE(vectorT) :: new

  !> make sure that the two vectors have the same number of entries
  IF(vec_length(self).NE.vec_length(other)) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  new = vec_add(self,other)

END FUNCTION vec_add_exc_


! ------------------------------------------------------------------------
!>  \brief Add a vectorT to a vectorT in place with mapping (non-exception checking)
!!
!!  \param self   Reference to the target vector object
!!  \param other  Reference to the input vector object
!!  \param ind    Integer mapping indices
!!
!!  This routine adds a vector to another vector in place according to an
!!  indexed mapping. That is, the element at (i) in other is added to the
!!  element at (ind(i)) of self.
! ------------------------------------------------------------------------
SUBROUTINE vec_mapped_add_ (self, other,ind)
  TYPE(vectorT), INTENT(INOUT) :: self
  TYPE(vectorT), INTENT(IN) :: other
  INTEGER, INTENT(IN) :: ind(:)
  INTEGER :: i    !> loop variable

  !> Add vector entries corresponding to mapping indices
  DO i = 1,SIZE(ind)

    IF (ind(i).EQ.0)  CYCLE

    CALL vec_set(self, ind(i), vec_get(self,ind(i)) + vec_get(other,i))

  END DO

END SUBROUTINE vec_mapped_add_


! ------------------------------------------------------------------------
!>  \brief Add a vectorT to a vectorT in place with mapping (exception checking)
!!
!!  \param self   Reference to the target vector object
!!  \param other  Reference to the input vector object
!!  \param ind    Integer mapping indices
!!  \param exc    Error code
!!
!!  This routine adds a vector to another vector in place according to an
!!  indexed mapping. That is, the element at (i) in other is added to the
!!  element at (ind(i)) of self. If the number of mapping indices does not
!!  match the size of other, it returns a DIMEN exception. If any of the
!!  indices indicate a location outside the vector, it returns a POSIT
!!  exception. Note that a zero index is valid and ind(i)=0 indicates that
!!  the value at (i) in the vector object will not be added to self.
! ------------------------------------------------------------------------
SUBROUTINE vec_mapped_add_exc_ (self, other,ind, exc)
  TYPE(vectorT), INTENT(INOUT) :: self
  TYPE(vectorT), INTENT(IN) :: other
  INTEGER, INTENT(IN) :: ind(:)
  INTEGER, INTENT(OUT) :: exc

  !> ensure that the number of indices matches the size of other
  IF (vec_length(other).NE.SIZE(ind)) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    RETURN

  !> ensure that there are no mapping indices outside of self
  ELSEIF ( MAXVAL(ind).GT.vec_length(self) .OR. MINVAL(ind).LT.0 ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL vec_mappedAdd(self,other,ind)

END SUBROUTINE vec_mapped_add_exc_


! ------------------------------------------------------------------------
!>  \brief Scalar multiplication (self*k version)
!!
!!  \param  self  Reference to the input vector object
!!  \param  k     Scalar multiplication factor
!!  \return new   Reference to the resulting vector object
!!
!!  This routine multiplies a vector by a scalar factor.
! ------------------------------------------------------------------------
FUNCTION vec_scal_mul_vec_scal_ (self, k) RESULT(new)
  TYPE(vectorT), INTENT(IN) :: self
  DOUBLE PRECISION, INTENT(IN) :: k
  TYPE(vectorT) :: new

  !> initialize solution vector
  CALL vec_init(new, vec_length(self))

  !> perform multiplication
  new%dat = k * self%dat

END FUNCTION vec_scal_mul_vec_scal_


! ------------------------------------------------------------------------
!>  \brief Scalar multiplication (k*self version)
!!
!!  \param  self  Reference to the input vector object
!!  \param  k     Scalar multiplication factor
!!  \return new   Reference to the resulting vector object
!!
!!  This routine multiplies a vector by a scalar factor. The
!!  overload is necessary for the OPERATOR (*) overload. This version
!!  simply calls the other version for better maintainability.
! ------------------------------------------------------------------------
FUNCTION vec_scal_mul_scal_vec_ (k, self) RESULT(new)
  TYPE(vectorT), INTENT(IN) :: self
  DOUBLE PRECISION, INTENT(IN) :: k
  TYPE(vectorT) :: new

  new = vec_scalMul(self,k)

END FUNCTION vec_scal_mul_scal_vec_


! ------------------------------------------------------------------------
!>  \brief Dot product of two vectors (non-exception checking)
!!
!!  \param  self  Reference to the first vector object
!!  \param  other Reference to the second vector object
!!  \return v     Value of the dot product
!!
!!  This routine takes the dot product of two vectors. That is, it takes
!!  the sum of the element-wise products of the entries in the vectors.
! ------------------------------------------------------------------------
FUNCTION vec_dot_prod_ (self,other) RESULT(v)
  TYPE(vectorT), INTENT(IN) :: self,other
  DOUBLE PRECISION :: v

  !> perform the dot product (using Fortran built-in for efficiency)
  v = DOT_PRODUCT(self%dat,other%dat)

END FUNCTION vec_dot_prod_


! ------------------------------------------------------------------------
!>  \brief Dot product of two vectors (exception checking)
!!
!!  \param  self  Reference to the first vector object
!!  \param  other Reference to the second vector object
!!  \return v     Value of the dot product
!!
!!  This routine takes the dot product of two vectors. That is, it takes
!!  the sum of the element-wise products of the entries in the vectors. If
!!  the length of the two vectors do not match, it returns a DIMEN
!!  exception.
! ------------------------------------------------------------------------
FUNCTION vec_dot_prod_exc_ (self,other, exc) RESULT(v)
  TYPE(vectorT), INTENT(IN) :: self,other
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: v

  !> ensure that the vectors have the same length
  IF (vec_length(self).NE.vec_length(other)) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    v=0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  v = vec_dotProd(self,other)

END FUNCTION vec_dot_prod_exc_

END MODULE vector_def
